# -*- coding: iso-8859-15 -*-
"""\
Datei:		ton.py
Zweck:		definiert Zugriff auf eine Klasse zur Tonerzeugung
Autor:		W. Arnhold
Erstellt:	07.05.2000
Aenderung:	umgeschrieben auf Floats 23.05.2003
Stand:		07.11.2003
"""

"""
  Dieses Modul definiert eine Klasse "Ton", die einen Ton als Schwingung
  von Energiewerten erzeugen kann. 

  Konstruktor:
  
  	Ton(dauer = 0.0, frequenz = 1000.0, lautstaerke = 1.0,
		phase = 0.0, tonfunktion = sinuston, rate = 22050)
	  Die Parameter bedeuten:
	  - dauer	: Tondauer in Sekunden (float)
	  - frequenz	: Frequenz in Hz (float)
	  - lautstaerke	: als Faktor zur Normallautstaerke (lautstaerke_1)
	  - phase	: Verschiebung im Verhltnis zu einer Schwingung:
	  		  0.0 = nicht verschoben, 0.5 = um eine halbe
			  Schwingung nach rechts verschoben, usw.
	  - tonfunktion	: Prozedur, s. u.
	  - rate	: Samples pro Sekunde (int) getestet mit 22050 und
			  44100 
	  Ohne Parameter wird ein Ton ohne Lnge erzeugt.

	  Eine selbstgeschriebene Tonfunktion bekommt einen Wert zwischen
	  0.0 und 1.0. Dies ist die relative Position innerhalb der
	  Schwingung. Sie liefert den Sampling-Wert im Bereich -1.0 bis 1.0.
	  Standard-Tonprozedur ist "sinuston", auerdem kann eine Prozedur
	  "stille" importiert werden, die nur Null-Samples liefert.

  Klassenprozeduren:
  
  	parameter()
	  liefert die aktuellen Werte Dauer, Frequenz, Lautstrke, Phase,
	  Samplingrate, Samplezahl und den absolut grten Sample-Wert.
	  Alles float, lediglich Samplingrate und Samplezahl int.
	  
  	samples()
	  liefert die Anzahl der Samples im Ton (int).
	
	sample(i)
	  liefert den Samplewert an Position i (float).
	
	anfuegen(float)
	  Der Parameter ist als neuer Sample-Wert ans Ende des Tons
	  angefuegt, alle anderen Werte sind nachgefuehrt.
	  
	null_sample(i)
	  liefert den Samplewert an Position i, wenn i ein legaler Index
	  ist, sonst 0 (float).

	sample_in_zeit(i)
	  liefert die Zeit, die bis zum Sample mit dem Index i vergangen
	  ist.
	  
	plot(datei, normiert = 0, zeitindex = 0)
	  "datei" ist eine zum Schreiben geffnete Datei, "normiert" und
	  "zeitindex" sind Wahrheitswerte. Schreibt die Samplewerte als
	  ASCII in die Datei, ein Paar aus Index und Wert je Zeile. Ist
	  "zeitindex" True, so ist der Index eine Gleitkommadarstellung der
	  Position des Samples in Sekunden, sonst die fortlaufende Nummer.
	  Ist "normiert" True, so ist der Samplewert im Bereich -1 .. +1
	  angegeben, sonst der wirdliche Wert. Die Datei muss extern
	  geffnet und geschlossen werden! Sie dient zum Inspizieren des
	  Tones oder besser als Eingabe fr ein Plotprogramm (z.B.  gnuplot).
	  
	+ (__add__(andererTon))
	  Additionsoperator liefert einen Ton, der aus den
	  aneinandergehngten beiden Tnen besteht, zwischen denen er
	  steht.

	[] (__getitem__(i) / __setitem__(i, samplewert))
	  Indexzugriff zum Lesen oder Schreiben von einzelnen Samplewerten.
	  Beispiel (t sei ein Ton):

	  	print t[123]
		t[123] = 0.98 * lautstaerke_1
		
	lautstaerke_setzen(faktor)
	  Verndert jedes Sample um den entsprechenden Lautstrkefaktor.
	  Keine Rckgabe.

	mischen(toene)
	  Voraussetzung: alle Tne haben gleiche Samplingrate.
	  "toene" ist entweder ein Paar, bestehend aus Einsatz und Ton oder
	  eine Liste solcher Paare (Tupel). Der Einsatz (float) gibt an, um
	  wieviel Sekunden nach (positiver Wert) oder vor (negativer Wert)
	  dem Ton, dessen Prozedur aufgerufen wird, der Ton des Tupels
	  eingemischt werden soll. Ist "toene" eine Liste, so werden alle
	  Tne eingemischt. Liefert den Mischton.

	einhuellen(huellkurve)
	  bergeben wird eine Instanz der Klasse "Huellkurve" (s.u.). Der
	  Ton ist auf der in der Hllkurve definierten Lnge durch diese
	  geformt.

	anwenden(func)
	  Auf jeden Samplewert wird die Funktion "func" angewandt und
	  der neue Samplewert zurckgeschrieben.

	speichern(datname)
	  Der Ton wird in eine Datei mit dem bergebenen Namen gespeichert.
	  Eine nicht vorhandene Endung ".ton" wird angehaengt.

	laden(datname)
	  Der Ton wird aus der angegebenen Datei geladen. Wenn der
	  Dateiname nicht die Endung ".ton" hat, wird diese ergnzt.
	  Die vorherigen Samples des Tones sind entfernt, alle Werte bis
	  auf die Tonfunktion sind nachgefhrt.

	wave_speichern(datname)
	  Der Ton wird als Wave-Datei geschrieben. Eine nicht vorhandene
	  Endung ".wav" wird angehaengt.

	wave_laden(datname)
	  Die Samples des Tones werden aus einer Wave-Datei geladen. Sie
	  werden entsprechend dem aktuellen Lautstrkewert angepasst. Die
	  Samplingrate und die Samplezahl werden bernommen, maximaler
	  absoluter Sample-Wert angepasst.

  Weiter wird eine Klasse Klang definiert, die Toene aus der Obertonreihe
  mischt, also solche, deren Frequenzen zueinander im Verhaeltnis 1:2:3:4
  usw. stehen.

  Konstruktor:

	Klang(dauer = 0.0, grundfrequenz = 1000.0, lautstaerke = 1.0,
	      obertoene = [], phase = 0.0, tonfunktion = sinuston,
	      rate = 22050)

	Die Parameter bedeuten das Gleiche wie bei Ton mit folgenden
	Ergaenzungen:

	- grundfrequenz	: Die Ausgangsfrequenz des untersten Tones
	- obertoene	: ist leer oder eine Liste von Paaren (Tupeln).

	Ist die Liste "obertoene" leer, so erhaelt man nur den Ton in der
	Grundfrequenz. Ansonsten bedeutet fuer jedes Tupel in der Liste
	der erste Wert eine Phasenverschiebung und der zweite die relative
	Lautstrke (wie bei Ton) des einzumischenden Obertones. Dessen
	Frequenzverhltnis zum Vorgaengerton folgt den Regeln der
	Obertonreihe, d.h. der erste Oberton hat zum Grundton ein
	Frequenzverhltnis von 2:1, der zweite Oberton zum ersten ein
	Verhltnis von 3:2, der dritte zum zweiten eines von 3:4 usw.

  Auerdem steht eine Klasse "Huellkurve" zur Verfgung, mit der man
  Hllkurven von Tnen beschreiben kann (s.o., Ton.einhuellen).
  Eine Hllkurve besteht aus vier Teilen:

  	- Attack-Teil :	Der Teil des ersten Anstiegs der Intentsitt bei
	  		Beginn des Tones.
	- Decay-Teil  :	Das Abschwellen auf die normale Intensitt.
	- Sustain-Teil:	Die gehaltene Dauerintensitt
	- Release-Teil:	Das Abschwellen der Tonintensitt bis zur Stille.

  Konstruktor:

	Huellkurve(l, a, d, r, s)

	Die Parameter bedeuten:

	- l : Lnge in Sekunden
	- a : Attack-Funktion
	- d : Decay-Funktion
	- s : Sustain-Funktion
	- r : Release-Funktion

	Die Parameter "a" bis "r" sind Paare, deren erster Wert den
	relativen Anteil an der Gesamtlnge der Hllkurve angibt, der
	zweite ist die Tonfunktion.

	Die Tonfunktionen erhalten einen Wert, der die relative Position
	innerhalb ihrer Kurvenphase angibt (z.B. fr Decay: 0.0 = am
	Anfang, 1.0 = am Ende der Decay-Phase) und die einen Wert zwischen
	0.0 und 1.0 zurckliefern.

  Zugriffe:

	faktor(zeitwert)
	  erhlt einen wert in Sekunden (float) und liefert den faktor
	  zurck, der zum multiplizieren des Sample-Wertes zu diesem
	  Zeitpunkt verwendet wird. Wird in Ton.einhuellen gebraucht.
"""
import array
import math
import wave

ARRAYTYP = 'd'			# Samples intern als "double" 
# Annahme: "Normale" Lautstrke ~ Parameterwert 1.0 ~ 50 Phon = 1.0E-11

lautstaerke_1 = 1.0E-11

# x = Bruchteil einer ganzen Schwingung, return: Sinuswert
def sinuston(x):
	return math.sin(x* 2 * math.pi)

def stille(x):
	return 0.0

def toenecmp(t1, t2):
	return cmp(t1[0], t2[0])
	
class Ton:
	def __init__(self,
			dauer = 0.0, 		# in Sekunden
			frequenz = 1000.0, 	# in Hz
			lautstaerke = 1.0, 	# 0.0 - 1.0 = lautstaerke_1
			phase = 0.0,		# Verschiebung
			tonfunktion = sinuston, # Tonerzeugungsfunktion
			rate = 22050):		# Samples pro Sekunde
		self.sub_init( dauer, frequenz, lautstaerke, phase,
				tonfunktion, rate)

		# Tonerzeugung
		s = 0
		for s in range(self.samplezahl):
			p = (self.phase + s * self.frequenz / self.rate) % 1
			wert = self.tonfunktion(p)  * lautstaerke_1 * \
				self.lautstaerke
			self._min_max_setzen(wert)
			self.reihe.append(wert)

	def sub_init(self, dauer, frequenz, lautstaerke, phase,
			tonfunktion, rate):
		self.dauer = float(dauer)
		self.frequenz = float(frequenz)
		self.lautstaerke = float(lautstaerke)
		self.phase = float(phase)
		self.tonfunktion = tonfunktion
		self.rate = rate
		self.inv_rate = 1.0 / self.rate
		self.samplezahl = int(self.dauer * rate)
		self.min_max = 0.0
		self.a_typ = ARRAYTYP
		self.reihe = array.array(ARRAYTYP)

	def parameter(self):
		return self.dauer, self.frequenz, self.lautstaerke, \
		self.phase, self.rate, len(self.reihe), self.min_max

	def samples(self):
		return len(self.reihe)

	def sample(self, i):
		return self.reihe[i]
	
	def __getitem__(self, i):
		return self.reihe[i]

	def _min_max_setzen(self, s):
		abs_s = abs(s)
		if abs_s > self.min_max:
			self.min_max = abs_s

	def __setitem__(self, i, s):
		self._min_max_setzen(s)
		self.reihe[i] = s

	def anfuegen(self, s):
		self._min_max_setzen(s)
		self.reihe.append(s)
		self.samplezahl = len(self.reihe)
		self.dauer = float(self.samplezahl) / self.rate 
		
	def null_sample(self, i):
		if 0 <= i < len(self.reihe):
			return self.reihe[i]
		return 0.0

	def sample_in_zeit(self, si):
		return si * self.inv_rate

	def plot(self, datei, normiert = 0, zeitindex = 0):
		for i in range(len(self.reihe)):
			s = self.reihe[i]
			if normiert:
				s = s / lautstaerke_1
				s = '%g' % s
			else:
				s = '%g' % s
			if zeitindex:
				ploti = '%.5f' % self.sample_in_zeit(i)
			else:
				ploti = '%d' % i
			datei.write("%s %s\n" % (ploti, s))

	def __add__(self, andererTon):
		t = Ton()
		t.rate = self.rate
		t.a_typ = self.a_typ
		t.min_max = max(self.min_max, andererTon.min_max)
		t.reihe = self.reihe + andererTon.reihe
		t.dauer = self.dauer + andererTon.dauer
		return t

	def lautstaerke_setzen(self, faktor = 1.0):	# Werte < 1: leiser, 
		for i in range(len(self.reihe)):	# Werte > 1: lauter
			w = self.reihe[i] * faktor
			self.reihe[i] = w
			self._min_max_setzen(w)

# "toene" ist 
# - ein Tupel aus einem Einsatz und einem Ton oder
# - eine Liste solcher Tupel.
# Einsatz ist ein Float in Sekunden, der angibt, zu welchem Zeitpunkt der
# Ton eingemischt werden soll. 
# - 0.0 ist zum gleichen Zeitpunkt
# - kleinere Werte : vorher,
# - groessere Werte: spaeter.

	def mischen(self, toene):
		if type(toene) == type([]):
			for t in toene:
				if not isinstance(t[1], Ton):
					raise ValueError, 'falscher Ton'
			tl = toene + [(0.0, self)]
		else:
			if not isinstance(toene[1], Ton):
				raise ValueError, 'falscher Ton'
			tl = [toene , (0.0, self)]
		tl.sort(toenecmp)
# Einsaetze in Sampleindizes umwandeln, kleinsten und Offset ermitteln:
		min_i = 0
		for i in range(len(tl)):
			t = tl[i]
			s_i = int(t[0] * self.rate + 0.5)
			offs = min_i - s_i
			tl[i] = (s_i, t[1])
			min_i = min(min_i, s_i)
# Offset fuer unterschiedlichen Einsatz bestimmen und einfuegen:
		for i in range(len(tl)):
			t = tl[i]
			tl[i] = (t[0], min_i - t[0], t[1])
# Gesamtlaenge bestimmen:
		gesl = 0l
		for t in tl:
			gesl = max(gesl, t[0] - min_i + t[2].samples())
		neue_reihe = array.array(self.a_typ)
		tonzahl = len(tl)
		neu_min_max = 0.0
		for i in range(gesl):
			w = 0.0
			for t in tl:
				w += t[2].null_sample(i + t[1])
			neue_reihe.append(w)
			abs_w = abs(w)
			if abs_w > neu_min_max:
				neu_min_max = abs_w
		self.reihe = neue_reihe
		self.min_max = neu_min_max

	def einhuellen(self, h):
		for i in range(len(self.reihe)):
			s = self.reihe[i]
			s *= h.faktor(self.sample_in_zeit(i))
			self.reihe[i] = s

	def anwenden(self, func):
		for i in range(len(self.reihe)):
			self.reihe[i] = func(i, self.reihe[i])

	def speichern(self, datname):
		if datname[-4:] != '.ton':
			datname += '.ton'
		dat = open(datname, 'wb')
		dat.write('%s\n' % `self.dauer`)
		dat.write('%s\n' % `self.frequenz`)
		dat.write('%s\n' % `self.lautstaerke`)
		dat.write('%s\n' % `self.phase`)
		dat.write('%s\n' % `self.rate`)
		dat.write('%s\n' % `self.samplezahl`)
		dat.write('%s\n' % `self.min_max`)
		dat.write('%s\n' % self.a_typ)
		self.reihe.tofile(dat)
		dat.close()

	def laden(self, datname):
		if datname[-4:] != '.ton':
			datname += '.ton'
		dat = open(datname, 'rb')
		self.dauer = eval(dat.readline()[:-1])
		self.frequenz = eval(dat.readline()[:-1])
		self.lautstaerke = eval(dat.readline()[:-1])
		self.phase = eval(dat.readline()[:-1])
		self.rate = eval(dat.readline()[:-1])
		self.inv_rate = 1.0 / self.rate
		self.samplezahl = eval(dat.readline()[:-1])
		self.min_max = eval(dat.readline()[:-1])
		self.a_typ = dat.readline()[0]
		if self.a_typ != ARRAYTYP:
			print 'Falscher Arraytyp', self.a_typ
			raise ValueError
		self.reihe = array.array(ARRAYTYP)
		self.reihe.fromfile(dat, self.samplezahl)

	def wave_speichern(self, datname):
                if datname[-4:] != '.wav':
                        datname += '.wav'
		if self.a_typ == 'B':
			samplewidth = 1
		else:
			samplewidth = 2
		wd = wave.open(datname, 'wb')
		wd.setparams((1, samplewidth, self.rate, len(self.reihe), 
				'NONE', 'not compressed'))
		lsbmaske = 255
		msbmaske = 255 << 8
		if self.min_max == 0.0:
			print 'Maximalwert ist Null!'
			wd.close()
			return
		faktor = 32767.0 / max(self.min_max, lautstaerke_1)
		for wert in self.reihe:
			try:
				sample = int(wert * faktor)
			except OverflowError, grund:
				print 'ton::Ton.wave_speichern:', grund,
				print 'fuer wert', wert, 'und faktor', faktor
				raise ValueError
			lsb = sample & lsbmaske
			msb = (sample & msbmaske) >> 8
			wd.writeframesraw(chr(lsb))
			wd.writeframesraw(chr(msb))
		wd.close()

	def wave_laden(self, datname):
                if datname[-4:] != '.wav':
                        datname += '.wav'
		wd = wave.open(datname, 'rb')
		if wd.getnchannels () != 1:
			raise ValueError, "nur Mono bitte!"
		self.rate = wd.getframerate ()
		self.a_typ = ARRAYTYP
		self.reihe = array.array(self.a_typ)
		n = wd.getnframes()
		print '%d samples gefunden in %s, rate: %d' % (n, datname, self.rate)
		faktor = lautstaerke_1 / 32767.0 * self.lautstaerke
		self.min_max = 0.0
		maske = 1 << 15
		for i in range(n):
			frame = wd.readframes(1)
			w = ord(frame[0]) + (ord(frame[1]) << 8)
			if w & maske:
				w -= 65536
			s = w * faktor
			self._min_max_setzen(s)
			self.reihe.append(s)
		self.dauer = float(len(self.reihe)) / self.rate 

class Klang(Ton):
	def __init__(self,
			dauer = 0.0,
			grundfrequenz = 1000.0,
			lautstaerke = 0.5,
			obertoene = [],
			phase = 0.0, 
			tonfunktion = sinuston,
			rate = 22050):
		if not obertoene:
			Ton.__init__(self, dauer, grundfrequenz,
				lautstaerke, phase, tonfunktion, rate)
			return
		self.sub_init( dauer, grundfrequenz, lautstaerke, phase,
				tonfunktion, rate)
		for s in range(self.samplezahl):
			f = self.frequenz
			p = (self.phase + s * f/ self.rate) % 1
			wert = self.tonfunktion(p) * self.lautstaerke
			for ti in range(len(obertoene)):
				f = f * (ti + 2.0) / (ti + 1.0)
				ophase, olautstaerke = obertoene[ti]
				p = (ophase  + s * f/ self.rate) % 1
				wert += self.tonfunktion(p) * \
					self.lautstaerke * \
					olautstaerke
			self.reihe.append(wert)
			self._min_max_setzen(wert)

class Huellkurve:
	def __init__(self,
			l,	# Laenge in s
			a,	# Attack-Teil,  Funktion
			d,	# Decay-Teil,   Funktion
			s,	# Sustain-Teil, Funktion
			r):	# Release-Teil, Funktion
		self.l = float(l)
		self.a = float(a[0]), a[1]
		self.grenze_a = self.l * self.a[0]
		self.d = float(d[0]), d[1]
		self.grenze_d = self.l * (self.a[0] + self.d[0])
		self.s = float(s[0]), s[1]
		self.grenze_s = self.l * (self.a[0] + self.d[0] +
						self.s[0])
		self.r = float(r[0]), r[1]
		self.breite_a = self.grenze_a
		self.breite_d = self.grenze_d - self.grenze_a
		self.breite_s = self.grenze_s - self.grenze_d
		self.breite_r = self.l - self.grenze_s
	
	def faktor(self, z):	# Zeitwert <= self.l
		zrel = z / self.l
		if z > self.grenze_s:
			return self.r[1]((z - self.grenze_s) / self.breite_r)
		elif z > self.grenze_d:
			return self.s[1]((z - self.grenze_d) / self.breite_s)
		elif z > self.grenze_a:
			return self.d[1]((z - self.grenze_a) / self.breite_d)
		else:
			return self.a[1](z / self.breite_a)
